import asyncio
import json
import aiohttp
import requests

from rich import inspect
from rich.logging import RichHandler
import logging

logging.basicConfig(
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler()],
    level=logging.DEBUG
)

log = logging.getLogger("modpackmaker")

import argparse
import io
import os
import random
from simple_term_menu import TerminalMenu
from bs4 import BeautifulSoup
import sys
import time
import zipfile

from lock import process_modpack_config

game_version = ['1.12.2']

curseforge_url = 'https://addons-ecs.forgesvc.net/api/v2/addon/'
modpack_search_url = f'{curseforge_url}search?gameId=432&categoryId=4472&searchFilter='

def parse_args(parser):
    args = parser.parse_args()
    return args


def init_args():
    parser = argparse.ArgumentParser(
        description='Wolfpackmaker (modpackmaker.py) (https://woofmc.xyz)'
    )
    parser.add_argument('modpack', metavar='MODPACK_SLUG', type=str, help='Search indication for a CurseForge modpack.')
    return parser

parser = init_args()
args = parse_args(parser)

headers = {
    'User-Agent': 'Wolfpackmaker (https://woofmc.xyz)',
    'Accept-Encoding': None
}

async def resolve_mod(session, file):
    async with session.get(f"{curseforge_url}{file['projectID']}") as r:
        data_mod = await r.json()
    async with session.get(f"{curseforge_url}{file['projectID']}/file/{file['fileID']}") as r:
        data_file = await r.json()
    mod_yaml = f"- {data_mod['slug']}:\n"
    mod_yaml += f"    url: {data_file['downloadUrl']}\n"
    log.info(f"{data_mod['slug']}...")
    return mod_yaml

async def resolve_mods(session, files):
    tasks = []
    async_session = aiohttp.ClientSession()
    for file in files:
        task = asyncio.create_task(resolve_mod(async_session, file))
        tasks.append(task)
    future = await asyncio.gather(*tasks)
    await async_session.close()
    return ''.join(future)

def main():
    session = requests.Session()
    session.headers.update(headers)
    try:
        int(args.modpack)
        r = session.get(f"{curseforge_url}{args.modpack}")
        result = [json.loads(r.content)]
        if len(result) == 1:
            selection = result[0]
        else:
            sys.exit(log.error("No modpack found."))
    except ValueError:
        r = session.get(f"{modpack_search_url}{args.modpack}")
        r.status_code != 200 and sys.exit(log.critical(f"Code {r.status_code}: {r.content}"))
        result = json.loads(r.content)
        if len(result) > 1:
            log.info("Multiple modpack choices found. Select the modpack of choice below.")
            terminal_menu = TerminalMenu([m['name'] for m in result])
            menu_entry_index = terminal_menu.show()
            selection = result[menu_entry_index]
        elif len(result) == 1:
            selection = result[0]
        else:
            sys.exit(log.error(f"No modpack found using search query {args.modpack}."))
    files = [f for f in selection['latestFiles']]
    if len(files) > 1:
        log.info(f"Multiple files for {selection['name']}. Choose which one you would like.")
        terminal_menu = TerminalMenu([f['displayName'] + ' (released on ' + f['fileDate'] + ')' for f in files])
        menu_entry_index = terminal_menu.show()
        file_selection = files[menu_entry_index]
    elif len(files) == 1:
        file_selection = files[0]
    else:
        sys.exit(log.error("No file found... somehow?"))
    zip_dir = f"modpack_data/{selection['slug']}"
    if not os.path.exists(zip_dir):
        modpack_file = io.BytesIO()
        with requests.get(file_selection['downloadUrl'], stream=True) as r:
            print(f"Downloading {file_selection['downloadUrl']}... ({r.headers['content-length']} bytes)")
            for chunk in r.iter_content(65535):
                modpack_file.write(chunk)
        zip = zipfile.ZipFile(modpack_file)
        zip.extractall(zip_dir)
        zip.close()
        modpack_file.close()
    manifest_yaml = str()
    with open(f"{zip_dir}/manifest.json") as f:
        cf_modpack_manifest = json.loads(f.read())
    inspect(cf_modpack_manifest)
    if os.path.exists(f"{zip_dir}/manifest.yml"):
        with open(f"{zip_dir}/manifest.yml") as f:
            manifest_yaml = f.read()
    else:
        if 'forge' in cf_modpack_manifest['minecraft']['modLoaders'][0]['id']:
            manifest_yaml += 'modloader: Forge\n'
        elif 'fabric' in cf_modpack_manifest['minecraft']['modLoaders'][0]['id']:
            manifest_yaml += 'modloader: Fabric\n'
        manifest_yaml += f'version: {cf_modpack_manifest["minecraft"]["version"]}\n'
        manifest_yaml += f'mods:\n'
        log.info("Resolving mods...")
        mods_yaml = asyncio.run(resolve_mods(session, cf_modpack_manifest['files']))
        manifest_yaml += mods_yaml
    if '- optifine:' in manifest_yaml.splitlines():
        do_optifine = False
    else:
        do_optifine = input("> Add OptiFine? (y/n) ").lower() == 'y'
    if do_optifine:
        optifine_url = "https://get.kalka.io/Optifine"
        log.info("Looking for OptiFine...")
        optifine_r = session.get(optifine_url)
        optifine_html = BeautifulSoup(optifine_r.content)
        for href in optifine_html.find_all('a'):
            optifine_link = href.get('href')
            if file_selection['gameVersion'][0] in optifine_link:
                break
        manifest_yaml += '- optifine:\n'
        manifest_yaml += f'    url: {optifine_url}{optifine_link}'
    log.info("Saving manifest.yml...")
    log.info(manifest_yaml)
    with open(f"{zip_dir}/manifest.yml", "w") as f:
        f.write(manifest_yaml)
    log.info("Locking mods...")
    modpack_lock = asyncio.run(process_modpack_config(manifest=manifest_yaml))
    with open(f"{zip_dir}/manifest.lock", "w") as f:
        f.write(json.dumps(modpack_lock))
    log.info("Done.")


main()



import os
import shutil
from os.path import exists, join

from tinydb import TinyDB, Query


class MultiMC:
    def __init__(self, wolfpackinst):
        wfi = wolfpackinst
        if exists(join(wfi.current_dir, "mmc-pack.json")):
            valid = True
        else:
            valid = False

        if not valid:
            wfi.log.warn(f"{wfi.current_dir} does not have a valid MultiMC json file. Ensure this is the directory you "
                         f"are intending to use with MultiMC.")
            return

        if wfi.args.dir:
            confirm = input(f"Are you sure you want to install {' '.join(wfi.modpack_name.split('-')).title()} into {wfi.current_dir}? (y/n): ")
            if confirm == 'y':
                pass
            elif confirm == 'n':
                return
            else:
                return

        multimc_dir = join(wfi.current_dir, ".minecraft")

        print(os.listdir(wfi.minecraft_dir))

        if wfi.needs_update:
            wfi.log.info("Copying MMC json file...")

            with open(join(wfi.minecraft_dir, "mmc-pack.json")) as f, open(join(wfi.current_dir, "mmc-pack.json"),
                                                                           'w') as fh:
                fh.write(f.read())

            wfi.log.info("Copying config...")

            shutil.copytree(join(wfi.minecraft_dir, "config"), join(multimc_dir, "config"), dirs_exist_ok=True)

        db = wfi.DB(f"{wfi.minecraft_dir}/history.json")

        modpack_data_cache = db.all()

        for cache in modpack_data_cache:
            if str(cache['id']) == str(wfi.new_version):
                for mod in cache['data']['mod_data']['mods']:
                    if not exists(join(multimc_dir, "mods", mod['filename'])):
                        with open(join(wfi.cached_dir, "mods", mod['filename']), 'rb') as f, open(join(multimc_dir, "mods", mod['filename']), 'wb') as fh:
                            fh.write(f.read())
            else:
                for mod in cache['data']['mod_data']['mods']:
                    if not exists(join(multimc_dir, "mods", mod['filename'])):
                        wfi.log.debug(f"Removing previous version {mod['filename']}")
                        os.remove(join(multimc_dir, "mods", mod['filename']))


        # TODO: Copy MultiMC json from cache dir
        # TODO: Check for updates and delete old mod files from mods folder






from posixpath import basename, dirname
from rich.console import Console
import logging
import time
from rich import inspect
from appdirs import user_cache_dir
import datetime
from dateutil import tz
from os import mkdir
import random

import owoify
from pyfiglet import Figlet



class Log(Console):
    log_contents = {}
    f = Figlet()

    def warn(self, msg):
        logging.root.level <= logging.WARN and self.log(f"[yellow][WARN][white] {msg}")
    def warning(self, msg):
        self.warn(msg)
    def info(self, msg):
        logging.root.level <= logging.INFO and self.log(f"[cyan][INFO][white] {msg}")
    def debug(self, msg):
        logging.root.level <= logging.DEBUG and self.log(f"[green][DEBUG][white] {msg}")
    def critical(self, msg):
        logging.root.level <= logging.CRITICAL and self.log(f"[red][CRITICAL][white] {msg}")

    def log(self, msg):
        super().log(msg)
        t = time.time()
        self.log_contents[t] = msg
    
    def parse_log(self, args):
        if args.verbose:
            logging.root.level = logging.DEBUG
            self.debug("Enabling verbose mode.")
        else:
            logging.root.level = logging.INFO

    def fancy_intro(self, description=''):
        self.info(str('').join(['####' for _ in range(16)]))
        self.info(self.f.renderText(("woofmc.xyz")))
        keywords = random.choice(
            ['A custom made Minecraft modpack script. Nothing special, hehe.',
            'Please don\'t tell anyone about this...',
            'Hehe. UwU, It\'s all we have, I know. I\'m sorry!',
            'I should probably get a better idea for this list...',
            'Not sponsored by Awoos!'
            ]
        )
        self.info(owoify.owoify(keywords))
        self.info(description)
        self.info(str('').join(['####' for _ in range(16)]))

    def save_log(self, module):
        self.log_cache = user_cache_dir(f'wolfpackmaker/log/{module}')
        try:
            mkdir(self.log_cache)
        except FileNotFoundError:
            try:
                mkdir(dirname(self.log_cache))
            except FileNotFoundError:
                mkdir(dirname(dirname(self.log_cache)))
        except FileExistsError:
            pass
        finally:
            try:
                mkdir(self.log_cache)
            except FileNotFoundError as e:
                self.warn(f"Could not save logfile {self.log_cache} with the error {e}.")
                return
            except FileExistsError:
                pass
        self.debug("Saving log...")
        tzinfo = tz.gettz('EDT')
        logfile_time = datetime.datetime.fromtimestamp(time.time(), tz=tzinfo)
        logfile_date = datetime.datetime.strftime(logfile_time, "%m_%d_%Y-%H_%M_%S")
        with open(f"{self.log_cache}/{logfile_date}.log", "w") as f:
            for k, v in self.log_contents.items():
                format_time = datetime.datetime.fromtimestamp(k, tz=tzinfo)
                format_date = datetime.datetime.strftime(format_time, "%m-%d-%Y-%H:%M:%S")
                f.write(f"[{format_date}] {v}\n")




import aiohttp
import argparse
import asyncio
import json
from rich.traceback import install as init_traceback
from util import Log


def parse_args(parser):
    args = parser.parse_args()
    return args


def init_args():
    parser = argparse.ArgumentParser(
        description='Wolfpackmaker (curseforgedb.py) (https://woofmc.xyz)'
    )
    parser.add_argument('--with-figlet', help='Defaults to True. Use Figlet when printing the wolfpackmaker intro')
    parser.add_argument("-v", "--verbose", action="store_true",
                            help="increase output verbosity")
    return parser


parser = init_args()
args = parse_args(parser)

headers = {'User-Agent':'wolfpackmaker (made by Kalka) business inquiries: b@kalka.io'}


async def get_curseforge_api(session, index, page_size, log, version=None):
    curseforge_url = f'https://addons-ecs.forgesvc.net/api/v2/addon/search?categoryId=0&gameId=432{version and "&gameVersion=" + str(version)}&sectionId=6&searchFilter=&sort=0'
    async with session.get(curseforge_url + '&pageSize={}&index={}'.format(page_size, index)) as r:
        data = await r.json()
        log.debug("Requested CurseForge API starting from index {}{}.".format(index,
                                                                                 version and ' for version ' + version or ''))
        return data

versions = [
    "1.16.5",
    "1.16.4",
    "1.16.3",
    "1.16.2",
    "1.16.1",
    "1.16",
    "1.12.2",
    "1.7.10"
]


async def process_curseforge_db(log):
    assert isinstance(log, Log)
    index = 0
    page_size = 50
    session = aiohttp.ClientSession()
    workers = []
    mods = []
    for i in range(200):
        workers.append(asyncio.create_task(get_curseforge_api(session, index, page_size, log)))
        index += page_size
    for v in versions:
        index = 0
        page_size = 50
        for i in range(200):
            workers.append(asyncio.create_task(get_curseforge_api(session, index, page_size, log, version=v)))
            index += page_size
    future = asyncio.gather(*workers)
    for d in await future:
        for m in d:
            if m['id'] in [v['id'] for v in mods]:
                continue
            mods.append({
                "id": m.get("id"),
                "name": m.get("name"),
                "summary": m.get("summary"),
                "slug": m.get("slug"),
                "latest_files": m.get("gameVersionLatestFiles") or m.get("latest_files") or m.get("latestFiles")
            })
    log.info(f"{len(mods)} mods.")
    await session.close()
    with open('curseforge.json', 'w') as f:
        log.debug("Saving mod data...")
        f.write(json.dumps(mods, indent=2))


def main():
    init_traceback()
    log = Log()
    log.parse_log(args)
    log.fancy_intro("Wolfpackmaker / curseforgedb.py")
    loop = asyncio.get_event_loop()
    loop.run_until_complete(process_curseforge_db(log))


if __name__ == '__main__':
    main()



import io
import shutil
from os import getcwd, remove
from os.path import dirname, join, exists
from pathlib import Path
from zipfile import ZipFile

from appdirs import user_cache_dir
from pkg_resources import get_distribution
from rich.progress import TextColumn, TransferSpeedColumn, DownloadColumn, BarColumn, TimeRemainingColumn, Progress
from rich.table import Column
from wolfpackmaker.launchers.multimc import MultiMC

from wolfpackutil import WolfpackUtil


class WolfpackMaker(WolfpackUtil):
    launchers = [
        "multimc"
    ]

    def __init__(self):
        super().__init__()
        self.__version__ = get_distribution('wolfpackmaker').version

        self.info = f"Wolfpackmaker {self.__version__}".split()
        self.name = self.info[0].lower()
        self.version = self.info[1]
        self.parser = self.init_args()
        self.args = self.parser.parse_args()
        self.log.parse_log(self.args)
        self.log.fancy_intro(self.name, version=self.version)
        self.repo_info = {
            "user": self.author,
            "repo": self.args.repo,
            "github_files": ['manifest.lock', 'config.zip']
        }

        self.modpack_name = self.args.repo.lower()

        self.headers = {
            'Authorization': 'token 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',
            'User-Agent': f'{self.name} {self.version}'
        }

        self.current_dir = self.args.dir and self.args.dir or getcwd()
        self.parent_dir = dirname(self.current_dir)
        self.cached_dir = user_cache_dir(self.name)

        self.modpacks_dir = join(self.cached_dir, 'modpacks')

        self.minecraft_dir = join(self.modpacks_dir, self.modpack_name)

        self.resourcepack_dir = join(self.minecraft_dir, 'resourcepacks')

        self.config_dir = join(self.minecraft_dir, 'config')

        self.mods_dir = \
            self.args.download and join(self.current_dir, self.args.download) or join(self.minecraft_dir, "mods")

        self.mods_cache_dir = \
            self.args.cache and join(self.current_dir, self.args.cache) or join(self.cached_dir, "mods")

        self.version = None
        self.new_version = None
        self.needs_update = True

    def main(self):
        self.create_folders()
        with self.Session() as session:
            session.headers.update(self.headers)
            self.log.info("Getting GitHub data...")
            github_data = session.get(self.github_url.format(self.repo_info['user'], self.repo_info['repo']))
            self.needs_update = True

            if exists(f'{self.minecraft_dir}/VERSION'):
                with open(f'{self.minecraft_dir}/VERSION') as fh:
                    self.version = fh.read()

            modpack_data = github_data.json()[0]
            self.new_version = str(modpack_data['id'])

            if self.version == str(modpack_data['id']):
                self.log.info("Modpack is already up to date.")
                self.needs_update = False

            mod_data = {}

            for asset in modpack_data['assets']:
                if asset['name'] in self.repo_info['github_files']:
                    mod_data[asset['name']] = asset['browser_download_url']

            mod_data['manifest.lock'] = session.get(mod_data['manifest.lock']).json()

            if self.needs_update:
                mod_data['config.zip'] = session.get(mod_data['config.zip']).content
                self.log.debug("Converting config.zip to BytesIO...")
                mod_data['config.zip'] = io.BytesIO(mod_data['config.zip'])
                self.log.debug("Converting to zip...")
                mod_data['config.zip'] = ZipFile(mod_data['config.zip'])
                self.log.info("Extracting modpack config...")
                mod_data['config.zip'].extractall(self.minecraft_dir)
                mod_data['config.zip'] = None

                if exists(join(self.minecraft_dir, ".minecraft")):
                    shutil.copytree(join(self.minecraft_dir, ".minecraft"), join(self.minecraft_dir, "config"),
                                    dirs_exist_ok=True)
                    shutil.rmtree(join(self.minecraft_dir, ".minecraft"))

                configignore_dir = join(self.minecraft_dir, "config", "config", ".configignore")

                if exists(configignore_dir):
                    with open(configignore_dir) as fh:
                        for line in fh.read().splitlines():
                            if exists(join(configignore_dir, line)):
                                self.log.info(f"Ignoring {line}...")
                                remove(join(configignore_dir, line))

            filesize = 0

            to_download = []

            with Progress(refresh_per_second=60, expand=True) as p:
                task = p.add_task(description="\t\tVerifying mods...", total=len(mod_data['manifest.lock']['mods']))
                for mod in mod_data['manifest.lock']['mods']:
                    if exists(join(self.cached_dir, "mods", mod['filename'])):
                        try:
                            with open(join(self.cached_dir, "mods", mod['filename']), 'rb') as fh:
                                local_length = len(fh.read())
                                verified = local_length == mod['fileLength']
                        except FileNotFoundError:
                            verified = False

                        if not verified:
                            print("\t" +
                                  f"{mod['name']} requires a re-download. "
                                  f"{local_length} / {mod['fileLength']} byte mismatch")
                            to_download.append(mod)
                    else:
                        to_download.append(mod)

                    p.advance(task)

                if not exists(join(self.cached_dir, "mods", mod['filename'])):
                    filesize += mod['fileLength']

            sorted_mods = reversed(sorted(to_download, key=lambda x: x['fileLength']))

            for mod in sorted_mods:
                with Progress(
                        TextColumn("\t> [yellow][progress.description]{task.description}",
                                   table_column=Column(ratio=4)),
                        TransferSpeedColumn(table_column=Column(ratio=4)),
                        DownloadColumn(table_column=Column(ratio=2)),
                        BarColumn(bar_width=None, table_column=Column(ratio=2)),
                        "[progress.percentage]{task.percentage:>3.0f}%",
                        TimeRemainingColumn(),
                        refresh_per_second=60,
                        expand=True
                ) as p:
                    task = p.add_task(description=mod['name'], total=int(mod['fileLength']))
                    mod_r = session.get(mod['downloadUrl'], stream=True)

                    with open(join(self.cached_dir, "mods", mod['filename']), 'wb') as fh:
                        for chunk in mod_r.iter_content(65535):
                            fh.write(chunk)
                            p.advance(task, len(chunk))

                    p.update(task, description=f"[green]{mod['name']}")

        with open(f'{self.minecraft_dir}/VERSION', 'w') as fh:
            fh.write(self.new_version)

        db = self.DB(f"{self.minecraft_dir}/history.json")

        if not db.contains(db.q.id == modpack_data['id']):
            db.insert({"id": modpack_data['id'],
                       "data": {"github_data": modpack_data, "mod_data": mod_data['manifest.lock']}})

        if self.args.launcher is None:
            self.log.warn(f"Please choose a launcher. Available launchers: {[launcher for launcher in self.launchers]}")
            return

        if self.args.launcher not in self.launchers:
            self.log.warn(
                f"{self.args.launcher} is not supported at this time. Available launchers: {[launcher for launcher in self.launchers]}")
            return

        if self.args.launcher == 'multimc':
            self.log.info("Assembling modpack with MultiMC...")
            MultiMC(self)

    def init_args(self):
        parser = self.Parser(description=f"{self.name} {self.version}", prog=f"{self.name}")
        parser.add_argument('-v', '--verbose', help='Increase output verbosity.', action='store_true')
        parser.add_argument('-r', '--repo', help="Wolfpack modpack repository.", required=True)
        parser.add_argument('--dir', help=f'Custom directory for Wolfpackmaker. Defaults to {dirname(getcwd())}')
        parser.add_argument('-mcdir', '--minecraft-dir', help='Specify custom minecraft dir. Defaults to .minecraft',
                            default='.minecraft')
        parser.add_argument('-l', '--launcher',
                            help=f"Specify the launcher. Available launchers: {[launcher for launcher in self.launchers]}")
        parser.add_argument('-d', '--download', help='Custom download directory')
        parser.add_argument('--cache', help='Custom cache directory')
        return parser

    def create_folders(self):
        Path(self.cached_dir).mkdir(parents=True, exist_ok=True)
        self.log.debug(f"Successfully created directory {self.cached_dir}")
        Path(self.minecraft_dir).mkdir(parents=True, exist_ok=True)
        self.log.debug(f"Successfully created directory {self.minecraft_dir}")
        Path(self.config_dir).mkdir(parents=True, exist_ok=True)
        self.log.debug(f"Successfully created directory {self.config_dir}")
        Path(self.mods_dir).mkdir(parents=True, exist_ok=True)
        self.log.debug(f"Successfully created directory {self.mods_dir}")
        Path(self.resourcepack_dir).mkdir(parents=True, exist_ok=True)
        self.log.debug(f"Successfully created directory {self.resourcepack_dir}")
        Path(self.mods_cache_dir).mkdir(parents=True, exist_ok=True)
        self.log.debug(f"Successfully created directory {self.mods_cache_dir}")
        Path(join(self.cached_dir, 'cached_config')).mkdir(parents=True, exist_ok=True)
        self.log.debug(f"Successfully created directory {join(self.cached_dir, 'cached_config')}")
        Path(self.modpacks_dir).mkdir(parents=True, exist_ok=True)
        self.log.debug(f"Successfully created directory {self.modpacks_dir}")



import io
import os
import aiohttp
import argparse
import asyncio
import datetime
import json
import logging
import sys
import time
import yaml


from aiohttp.client_exceptions import ContentTypeError
from collections import Counter
from os.path import basename
from rich.traceback import install as init_traceback
from util import Log

TYPE_FABRIC = 4
TYPE_FORGE = 1


log = Log()

found_mods = []
mod_slugs = []


def parse_args(parser):
    args = parser.parse_args()
    return args


def init_args():
    parser = argparse.ArgumentParser(
        description='Wolfpackmaker (lock.py) (https://woofmc.xyz)'
    )
    parser.add_argument('-v', '--verbose', help='Increase output verbosity.', action='store_true')
    parser.add_argument('-m', '--manifest', help='Optional location for the manifest e.g /opt/manifests/manifest.yml.'
                                                 '\nDefaults to workdir (manifest.yml)')
    parser.add_argument('--with-figlet', help='Defaults to True. Use Figlet when printing the wolfpackmaker intro')
    return parser

timeout = aiohttp.ClientTimeout(total=1)
retries = 5

async def fetch_file(curseforge_url, mod, session, fileId):
    files_url = f"{curseforge_url}{mod['id']}/file/{fileId}"
    for i in range(retries):
        try:
            async with session.get(files_url, timeout=timeout) as r:
                try:
                    file = await r.json()
                except ContentTypeError:
                    log.error(mod)
                    sys.exit(log.error("ContentType error."))
            break
        except asyncio.TimeoutError:
            log.info(f"Retrying {mod['name']} ({i+1} of {retries})...")
            continue
    return file


async def set_content_length(curseforge_url, session, mod_slug):
    async with session.get(curseforge_url) as r:
        content_length = int()
        try:
            content_length = int(r.headers['content-length'])
        except KeyError:
            log.warning("Could not get Content-Length directly, getting it ourselves...")
            async for chunk in r.content.iter_any():
                content_length += len(chunk)
            log.info(content_length)

    for m in found_mods:
        if mod_slug == m['slug']:
            log.info(f"Appended file length to {m['name']}")
            m.update({"fileLength": content_length})


async def fetch_mod(curseforge_url, mod_id, session):
    mod_url = curseforge_url + str(mod_id)
    for i in range(retries):
        try:
            async with session.get(mod_url, timeout=timeout) as r:
                # log.debug("Responding to request {}...".format(mod_url))
                try:
                    mod = await r.json()
                except ContentTypeError:
                    sys.exit(log.error("ContentType error."))
            break
        except asyncio.TimeoutError:
            log.info(f"Retrying {mod_id['name']}...")
            continue
    return mod


modloader = ''

import datetime


async def get_mod_file(curseforge_url, modpack_manifest, version, mc_version, mod, session, is_file_found):
    if is_file_found:
        return
    try:
        mod_compat = version["modLoader"]
    except KeyError:
        mod_compat = None
    
    if modpack_manifest["modloader"].lower() == 'forge' and mod_compat == TYPE_FABRIC:
        return
    if modpack_manifest["modloader"].lower() == 'fabric' and mod_compat == TYPE_FORGE:
        return
    if version["gameVersion"] in mc_version:
        file_id = version["projectFileId"]
        file = await fetch_file(curseforge_url, mod, session, file_id)
        return file


async def fetch_mod_data(curseforge_url, mod, session, modpack_manifest, cf_data, completed, to_complete):
    start_time = time.time()
    mc_version = [modpack_manifest["version"]]
    # if "1.16.5" in mc_version:
    #     for i in range(1, 5):
    #         mc_version.append(f"1.16.{i}")
    file_found = False

    try:
        latest_files = mod['latest_files']
    except KeyError:
        try:
            latest_files = mod['gameVersionLatestFiles']
        except KeyError:
            log.info(mod)
            latest_files = mod['latestFiles']
    
    for version in latest_files:
            file = await get_mod_file(curseforge_url, modpack_manifest, version, mc_version, mod, session, file_found)
            if not file: continue
            file_found = True
            deps = []
            for dep in file["dependencies"]:
                if dep["addonId"] in [m["id"] for m in found_mods]:
                    break
                deps += [d for d in cf_data if dep["addonId"] == d["id"] and dep["type"] == 3]
            dep_file_found = False
            for d in deps:
                if d['slug'] in [m for m in mod_slugs]:
                    break
                log.info(f"Resolving dependency {d['name']} for mod {mod['name']}...")
                mod_slugs.append(d['slug'])
                for df in d["latest_files"]:
                    dep_file = await get_mod_file(curseforge_url, modpack_manifest, df, mc_version, d, session, dep_file_found)
                    if not dep_file: continue
                    dep_file_found = True
                    found_mods.append({
                        "id": d["id"],
                        "slug": d["slug"],
                        "name": d["name"],
                        "downloadUrl": dep_file["downloadUrl"],
                        "filename": dep_file["fileName"],
                        "fileLength": dep_file["fileLength"]
                    })
            for m in found_mods:
                if m.get('downloadUrl') is not None:
                    continue
                if m["id"] == mod["id"]:
                    m.update({'downloadUrl': file['downloadUrl'], 'filename': file['fileName'], 'fileLength': file['fileLength']})
    completed[0] += 1
    log.info(f"[LOCK] [{completed[0]}/{to_complete[0]}] {mod['name']} took {time.time() - start_time:.3f} seconds.")
    if not file_found:
        log.warning(
            f"Mod {mod['slug']} [{mod['name']}] does not have an apparent version for {mc_version}, tread with caution")

minecraft_version = []

async def process_modpack_config(manifest):
    chunked = True  # Should chunk
    curseforge_url = 'https://addons-ecs.forgesvc.net/api/v2/addon/'
    modpack_manifest = yaml.load(manifest, Loader=yaml.SafeLoader)
    minecraft_version.append(modpack_manifest["version"])
    mods = modpack_manifest["mods"]
    # check for duplicates
    duplicate_mods = []
    for idx, mod in enumerate(mods):
        for k, v in mods[idx].items():
            mod_slugs.append(k)
            duplicate_mods.append(k)
    if [k for k,v in Counter(duplicate_mods).items() if v>1]:
        sys.exit(log.critical(f"Found duplicates in the manifest file. Please remove them before continuing:\n> {[k for k,v in Counter(duplicate_mods).items() if v>1]}"))
    curseforge_download_url = "https://vulpera.com/curseforge.json"
    session = aiohttp.ClientSession()
    log.debug(f"Established session {session}")
    log.info(f"Reading CurseForge data from {curseforge_download_url}")
    start_time = time.time()
    async with session.get(curseforge_download_url) as r:
        date = datetime.datetime.strptime(r.headers["last-modified"], "%a, %d %b %Y %H:%M:%S %Z")
        log.info(f"CurseForge DB date is {datetime.datetime.strftime(date, '%B %d, %Y at %H:%M:%Sz')}")
        if chunked:
            log.info("Reading chunked data... (it's probably big)")
            data = io.BytesIO()
            async for c in r.content.iter_chunked(65535):
                data.write(c)
            data.seek(0)
            curseforge_data = json.loads(data.read())
        else:
            data = await r.read()
            curseforge_data = json.loads(data)
    log.info(f"Took {time.time() - start_time:.2f}s. {len(curseforge_data)} mods recognized.")
    tasks = []
    to_complete = [0]
    completed = [0]
    for idx, mod in enumerate(mods):
        for k, v in mods[idx].items():
            client_only, server_only, optional = False, False, False
            match v:
                case {'clientonly': True}:
                    client_only = True
                case {'serveronly': True}:
                    server_only = True
                case {'optional': True}:
                    optional = True
            not_found_msg = f'This happened because we exhausted all efforts to search for {k}, and the only info we know about it is the mod slug, which is just {k}. The easiest fix to this is to visit https://www.curseforge.com/minecraft/mc-mods/{k} and copy the value of "Project ID", and append it to the corresponding mod in the yaml manifest, e.g:\n- {k}:\n    id: <id>... \nThe script will continue and disregard this specific mod, but it will be considered a mod we cannot digest!'
            finished_suffix = " (took {:.2f} seconds)"
            start_time
            try:
                mod_data = [m for m in curseforge_data if k == m['slug']][0]
            except IndexError:
                mod_data = [{
                    "id": None,
                    "name": k
                }][0]
            custom = [False]
            has_id = [False]
            match v:
                case {'id': id}:
                    try:
                        mod_data["filename"]
                    except KeyError:
                        to_complete[0] += 1
                        log.info(f"Using {id} for {k}. This should guarantee a positive match.")
                        cf_get = await session.get(curseforge_url + str(id))
                        data = await cf_get.json()
                        if k != data['slug']:
                            sys.exit(log.critical(f"Mod mismatch! {k} =/= {data['slug']}. This is usually impossible unless you are using the wrong mod ID."))
                        log.info(f"[MATCH] [{completed[0]}/{to_complete[0]}] Resolved {data['name']} through CurseForge!")
                        found_mods.append({
                            "id": data['id'],
                            "slug": data['slug'],
                            "name": data['name'],
                            "clientonly": client_only,
                            "serveronly": server_only,
                            "optional": optional
                        })
                        task = asyncio.create_task(fetch_mod_data(curseforge_url, data, session, modpack_manifest, curseforge_data, completed, to_complete))
                        tasks.append(task)
                        has_id[0] = True
                case {'url': url}:
                    task = asyncio.create_task(set_content_length(url, session, k))
                    tasks.append(task)
                    if url.split('.')[-1] == 'zip':
                        log.info(f"Handling resourcepack {k}...")
                        found_mods.append({
                            "id": mod_data['id'] or None,
                            "name": k,
                            "slug": k,
                            "filename": basename(url),
                            "downloadUrl": url,
                            "resourcepack": True
                        })
                        continue
                    found_mods.append({
                        "id": mod_data['id'] or None,
                        "name": mod_data['name'] or k,
                        "slug": k,
                        # the = is for optifine
                        "filename": '=' in url and url.split("=")[1] or basename(url),
                        "downloadUrl": url,
                        "clientonly": client_only,
                        "serveronly": server_only,
                        "optional": optional,
                        "custom": True,
                    })
                    to_complete[0] += 1
                    completed[0] += 1
                    log.info(f"[LOCK] [{completed[0]}/{to_complete[0]}] Resolved {mod_data['name']}, using custom URL {url}")
                    custom[0] = True
            if has_id[0]: continue
            if custom[0]: continue
            to_complete[0] += 1
            log.info(f"[MATCH] [{completed[0]}/{to_complete[0]}] Resolved {mod_data['name']} {finished_suffix.format(time.time() - start_time)}!")
            found_mods.append({
                "id": mod_data['id'] or None,
                "name": mod_data['name'] or k,
                "slug": k,
                "clientonly": client_only,
                "serveronly": server_only,
                "optional": optional
            })
            task = asyncio.create_task(fetch_mod_data(curseforge_url, mod_data, session, modpack_manifest, curseforge_data, completed, to_complete))
            tasks.append(task)
    await asyncio.gather(*tasks)
    await session.close()
    return found_mods


def save_lockfile():
    log.info("Saving lockfile...")
    with open('manifest.lock', 'w') as f:
        f.write(json.dumps({"version": minecraft_version[0], "mods": found_mods}))
    log.info("Saving pretty-printed file...")
    with open('manifest.json', 'w') as f:
        f.write(json.dumps({"version": minecraft_version[0], "mods": found_mods}, indent=2))

import requests

import sys

from rich import inspect

def main():
    init_traceback()
    parser = init_args()
    args = parse_args(parser)
    log.parse_log(args)
    log.fancy_intro(parser.description)
    loop = asyncio.new_event_loop()
    if args.manifest:
        if 'yml' and not 'https' in args.manifest:
            args.manifest = open(args.manifest).read()
        elif 'https' and 'yml' in args.manifest:
            args.manifest = requests.get(args.manifest).text
    else:
        if os.path.exists('manifest.yml'):
            with open('manifest.yml') as f:
                args.manifest = f.read()
    task = loop.create_task(process_modpack_config(manifest=args.manifest))
    loop.run_until_complete(task)
    save_lockfile()
    log.save_log("lock")
    sys.exit()


if __name__ == '__main__':
    main()
